# coding: utf-8
"""Converter to ingest scikit-learn models into Treelite"""
import warnings

warnings.warn(('treelite.gallery.sklearn has been moved to treelite.sklearn. ' +
               'treelite.gallery.sklearn will be removed in version 1.1.'),
              FutureWarning)
